from agador.utils import parse_command_map
from agador.nornir.logging import LOG_FORMAT, DEFAULT_LOGFILE
import logging
import argparse
import sys
from decouple import config
import subprocess

from datetime import datetime
import time

logger = logging.getLogger("agador")

LOOP_INTERVAL = 10



def main():
    parser = argparse.ArgumentParser(description="Run agador")
    parser.add_argument("-l", "--log-level", default=logging.INFO, help="Set log level for agador only")
    parser.add_argument("--echo", action="store_true", help="echo logfile to stdout")
    args = parser.parse_args()

    # setting up logging
    file_handler = logging.handlers.RotatingFileHandler(
        DEFAULT_LOGFILE, maxBytes=1024 * 1024 * 10, backupCount=20
    )
    file_handler.setFormatter(logging.Formatter(LOG_FORMAT))
    logger.addHandler(file_handler)
    logger.setLevel(args.log_level)
    if args.echo:
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setFormatter(logging.Formatter(LOG_FORMAT))
        logger.addHandler(stdout_handler)
    
    base_run_cmd = ['python', 'run.py', '-l', args.log_level, '--cmds']

    cmd_map = parse_command_map()
    current_run = None
    
    last_check = datetime.now()
    while True:

        # waiting a bit
        time.sleep(LOOP_INTERVAL)

        # checking the time
        now = datetime.now()

        # if any commands are scheduled to run between now and the last time
        # we checked the time, put them on a list of commands to run
        cmds_to_run = []
        for cmd, data in cmd_map.items():

            run_time = data["frequency"].schedule(now).prev()
            if run_time >= last_check:
                cmds_to_run.append(cmd)

        # If we have commands to run, first check to see if the previous run has completed.
        # if it has, run our new commands
        if cmds_to_run:
            if current_run and current_run.poll() is None:
                logger.error(f"Agador wants to run cmds {','.join(cmds_to_run)} but previous run (PID {current_run.pid}) has not completed!")
            else:
                cmd_args = base_run_cmd.copy()
                cmd_args.extend(cmds_to_run)
                current_run = subprocess.Popen(cmd_args)
                logger.info(f"Starting agador run (PID {current_run.pid}): {cmd_args}")

        last_check = now
        

if __name__ == "__main__":
    main()


import logging
import traceback
from nornir.core.task import Result, Task
from nornir.core.exceptions import NornirSubTaskError
from sqlalchemy.engine import Engine

from umnet_napalm import get_network_driver

from ..utils import get_device_cmd_list
from ..mappers.save_to_file import SaveResult
from ..mappers.save_to_db import ResultsToDb

from .nornir_setup import NUM_WORKERS
from decouple import config
from sqlalchemy import create_engine

logger = logging.getLogger(__name__)


def process_device(
    task: Task,
    cmd_list: list,
    cmd_map: dict,
) -> Result:
    """
    Run a set of commands against a specific device based on the
    command mapper and an optional restricted list of commands.
    If no list is provided all commands in the mapper will be run
    """
    failed = False
    db_engine = create_engine(
        config("DB_URL"), pool_size=NUM_WORKERS, max_overflow=NUM_WORKERS + 10
    )
    # first figure out which commands to run against this host
    cmd_list = get_device_cmd_list(cmd_map, task.host, cmd_list_filter=cmd_list)

    logger.info(f"Running getters on {task.host.name}: {cmd_list}")
    if not cmd_list:
        logger.info(f"No commands to execute for {task.host.name}")
        return Result(host=task.host, result="No commands to execute")

    # finding that for multiprocessing it's cleaner to use the context
    # manager - by this we mean 'with network_driver as device' - because
    # connection failures that raise exceptions aren't well tolerated.
    network_driver = get_network_driver(task.host.platform)
    parameters = {
        "hostname": task.host.hostname,
        "username": task.host.username,
        "password": task.host.password,
        "optional_args": {},
    }
    parameters.update(task.host.connection_options["umnet_napalm"].extras)

    getter_results = {}
    try:
        with network_driver(**parameters) as device:

            for cmd, options in cmd_list.items():
                method = getattr(device, cmd_map[cmd]["getter"])
                getter_results[cmd] = method(**options)

    except Exception as e:
        tb = traceback.format_exc()
        logger.error(
            "Host %r getters failed with traceback:\n%s",
            task.host.name,
            tb,
        )
        return Result(host=task.host, result=tb, error=str(e), failed=True)

    # for each command save data using the appropriate methods based
    # on the command map
    for cmd, result in getter_results.items():

        if "save_to_db" in cmd_map[cmd]:
            try:
                task.run(
                    name=f"{cmd}_save_to_db",
                    task=update_table,
                    result=result,
                    engine=db_engine,
                    mapper=cmd_map[cmd]["save_to_db"],
                )
            except NornirSubTaskError:
                failed = True

        if "save_to_file" in cmd_map[cmd]:
            try:
                task.run(
                    name=f"{cmd}_save_to_file",
                    task=save_to_file,
                    result=result,
                    repo_path=cmd_map[cmd]["save_to_file"]["destination"],
                    mapper=cmd_map[cmd]["save_to_file"]["mapper"],
                )
            except NornirSubTaskError:
                failed = True

    return Result(host=task.host, result="Completed", failed=failed)


def update_table(
    task: Task,
    result: Result,
    engine: Engine,
    mapper: ResultsToDb,
) -> Result:
    """
    Generic task for taking the results from a previous "napalm_get"
    task and saving it to the database
    """
    mapper(task.host.name, task.host.hostname, result, engine)

    return Result(host=task.host)


def save_to_file(
    task: Task,
    result: dict,
    repo_path: str,
    mapper: SaveResult,
) -> Result:
    """
    Task for saving to a file
    """
    mapper(task.host, repo_path, result)

    return Result(host=task.host)



from nornir.core import Nornir
from .credentials import CredentialMap


def update_nornir_credentials(nr: Nornir, cyberark_env_file: str):
    """
    Uses the credential map to update host credentials in this nornir instance
    """
    cred_map = CredentialMap(cyberark_env_file)
    cred_map.set_default_credentials(nr)
    for host in nr.inventory.hosts.values():
        cred_map.set_host_credentials(host)



import re
from typing import Union
from dataclasses import dataclass
import logging

from nornir.core import Nornir
from nornir.core.inventory import Host, ConnectionOptions
import yaml

from .cyberark import Cyberark, CyberarkLookupError

logger = logging.getLogger(__name__)


@dataclass
class CustomCredential:
    """
    Class that holds data around a custom credential stored in cyberark.
    """

    _c: Cyberark
    host_match: re.Pattern
    env: str
    username: str
    password_host: str

    # cached credentials for this object
    _user_password = False
    _user_enable = False

    def get_user_password(self) -> Union[str, None]:
        """
        Attempts to look up the password tied to this username and 'host' in cyberark.
        Returns None if no password was found
        """
        if self._user_password is False:
            try:
                self._user_password = self._c.lookup_username_password(
                    self.username, host=self.password_host, environment=self.env
                )
            except CyberarkLookupError:
                self._user_password = None

        return self._user_password

    def get_user_enable(self) -> Union[str, None]:
        """
        Returns enable secret after loooking it up in cyberark
        """
        if self._user_enable is False:
            try:
                self._user_enable = self._c.lookup_enable(
                    host=self.password_host, environment=self.env
                )
            except CyberarkLookupError:
                self._user_enable = None

        return self._user_enable


class CredentialMap:
    """
    Class that stores custom credentials defined in the
    credential map and can look up credentials in cyberark
    """

    def __init__(
        self,
        cyberark_env_file: str,
        map_file: str = "./agador/credentials/credential_map.yml",
    ):
        """
        Parses credential map file into list of credential objects
        """

        with open(map_file, encoding="utf-8") as fh:
            credential_map = yaml.load(fh, yaml.Loader)

        self.default_user = None
        self.default_enable = None
        self.default_cyberark_env = None

        # three default values are required in the credential map file
        for default in ["default_user", "default_enable", "default_cyberark_env"]:
            if default not in credential_map:
                raise ValueError(f"{default} must be set in credential map!")
            setattr(self, default, credential_map[default])

        self.cyberark = Cyberark(cyberark_env_file)

        # parsing each custom value
        self.custom = []
        for custom in credential_map["custom"]:

            # host match is required
            if not custom.get("host_match"):
                raise ValueError(
                    f"No host match specified for custom credential {custom}"
                )

            # and it must be a valid regex string
            try:
                host_match = re.compile(custom["host_match"])
            except TypeError as e:
                raise ValueError(
                    f"Uncompileable host match {custom['host_match']}"
                ) from e

            self.custom.append(
                CustomCredential(
                    _c=self.cyberark,
                    host_match=host_match,
                    env=custom.get("env", self.default_cyberark_env),
                    username=custom.get("username"),
                    password_host=custom.get("password_host"),
                )
            )

    def set_default_credentials(self, nr: Nornir):
        """
        Sets default username/password for nornir
        """
        nr.inventory.defaults.username = self.default_user
        nr.inventory.defaults.password = self.cyberark.lookup_username_password(
            self.default_user
        )

    def set_host_credentials(self, host: Host):
        """
        Sets the host credentials of a given nornir object if a custom credential is
        found for it. If there isn't a custom credential in the credential map
        it's assumed the default credentials are set and those are sufficient
        """

        for cred in self.custom:
            if re.match(cred.host_match, host.name):

                custom_password = cred.get_user_password()
                if custom_password:
                    host.username = cred.username
                    host.password = custom_password
                    logger.debug(
                        "Setting custom user/pass for %s: %s %s",
                        host.name,
                        host.username,
                        host.password,
                    )
                else:
                    logger.error(
                        "No password found in cyberark for %s and cred_map match %s",
                        host.name,
                        cred.host_match,
                    )

                # in general we expect a corresponding custom enable if there's a custom password
                # but that's not always the case
                custom_enable = cred.get_user_enable()
                if custom_enable:
                    conn_options = ConnectionOptions(
                        extras={"optional_args": {"secret": custom_enable}}
                    )
                    host.connection_options["umnet_napalm"] = conn_options
                    logger.debug(
                        "Setting custom secret for %s: %s", host.name, custom_enable
                    )

                return



from typing import Optional, Dict
import ipaddress
import sys
from os import getenv, getcwd
from os.path import isdir
import re
import argparse
from git import Repo
import yaml
from decouple import config, UndefinedValueError
from cron_converter import Cron

from nornir.core.inventory import Host

from umnet_napalm.abstract_base import AbstractUMnetNapalm

from .nornir import inventory_filters
from .mappers import save_to_db, save_to_file

class CommandMapError(Exception):
    def __init__(self, cmd:str, error_str:str):
        super().__init__(f"command map error for '{cmd}' - {error_str}")

def get_from_args_or_env(cli_arg: str, args: argparse.Namespace) -> str:
    """
    Pull a value from parsed arparse, if it's not there look for it
    in the environment
    """
    if getattr(args, cli_arg, False):
        return args.arg

    env_arg = cli_arg.replace("-", "_").upper()
    if getenv(env_arg):
        return getenv(env_arg)

    print(
        f"ERROR: Please provide {cli_arg} as cli input or as {env_arg} environment variable"
    )
    sys.exit(1)


def is_ip_address(ip: str) -> bool:
    """
    Returns whether a string is an IP (eg 10.233.0.10 or fe80::1)
    """
    try:
        ipaddress.ip_address(ip)
    except ValueError:
        return False
    return True


def is_ip_or_prefix(ip: str) -> bool:
    """
    Returns whether a string is an ip (10.233.0.10) or
    IP + prefix (10.233.0.10/24). IPv6 works as well.
    """
    try:
        ipaddress.ip_interface(ip)
    except ValueError:
        return False
    return True


def git_update(repo_path: str, commit_message: str):
    """
    Updates git repo and if applicable, the remote origin.
    Will overwrite origin.
    """
    repo = Repo(repo_path)
    status = repo.git.status()

    repo.git.add(all=True)
    repo.index.commit(commit_message)
    try:
        origin = repo.remote(name="origin")
    except ValueError:
        origin = None

    if origin:
        origin.push()


def parse_command_map() -> dict:
    """
    Parses and validates command_map file, replacing text references
    to functions to references to the actual functions where applicable.

    Future task: pydantic-based validation
    """

    db_module = save_to_db
    file_module = save_to_file

    with open(f"{getcwd()}/agador/command_map.yml", encoding="utf-8") as fh:
        cmd_map = yaml.safe_load(fh)

    output = {}
    for cmd, data in cmd_map["commands"].items():

        output[cmd] = {}

        # validating frequency, which is required
        if "frequency" not in data:
            raise CommandMapError(cmd, "Must specify frequency")
        try:
            output[cmd]["frequency"] = Cron(data["frequency"])
        except ValueError:
            raise CommandMapError(cmd, "Invalid frequency - must be in crontab format")

        # validating umnet_napalm getter specification
        if "getter" not in data:
            raise CommandMapError(cmd,"Must specify umnet_napalm getter")
        if data["getter"] not in dir(AbstractUMnetNapalm):
           raise CommandMapError(cmd, f"Unknown umnet_napalm getter {data['getter']}")
        output[cmd]["getter"] = data["getter"]

        # validating and retrieving inventory filter fuction
        inv_filter = data.get("inventory_filter", None)
        if inv_filter:
            if inv_filter not in dir(inventory_filters):
                raise CommandMapError(cmd, f"Unknown inventory filter {inv_filter}")

            output[cmd]["inventory_filter"] = getattr(inventory_filters, inv_filter)

        # validating and retrieving save_to_file class
        file_data = data.get("save_to_file", None)
        if file_data:
            if "mapper" not in file_data:
                raise CommandMapError(cmd, "Must specify mapper for save_to_file")

            if "destination" not in file_data:
                raise CommandMapError(cmd, "Must specify destination for save_to_file")

            destination = resolve_envs(file_data["destination"])
            if not isdir(destination):
                raise CommandMapError(cmd, "Invalid desintation folder for save_to_file")
            
            if file_data["mapper"] not in dir(file_module):
                raise CommandMapError(cmd, f"Unknown save_to_file mapper {file_data['mapper']}")

            output[cmd]["save_to_file"] = {
                "mapper": getattr(file_module, file_data["mapper"]),
                "destination": destination,
            }

        # validating and retrieving save_to_db class
        db_data = data.get("save_to_db", None)
        if db_data:
            if db_data not in dir(db_module):
                raise CommandMapError(cmd, f"Unknown save_to_db mapper {db_data}")

            output[cmd]["save_to_db"] = getattr(db_module, db_data)

        if not db_data and not file_data:
            raise CommandMapError(cmd,"Must specifiy either save_to_db or save_to_file")

    return output


def get_device_cmd_list(
    cmd_map: dict, host: Host, cmd_list_filter: Optional[list] = None
) -> Dict[str, dict]:
    """
    Gets list of commands tied to a device based on its
    host inventory data. Optionally provide a list of commands to
    restrict the output to.

    Returns a dict mapping the getter commands to run to the arguments they need to be run
    with (if any)
    """
    cmd_list = {}
    for cmd, data in cmd_map.items():

        if cmd_list_filter and cmd not in cmd_list_filter:
            continue

        if not data.get("inventory_filter") or data["inventory_filter"](host):
            cmd_list[cmd] = data.get("getter_args", {})

    return cmd_list

def resolve_envs(input_str:str)->str:
    """
    Takes an input string and searches for all instances of '${ENV_VAR}', replacing
    ENV_VAR with the value in the .env file. Raises an exception
    if the ENV_VAR is not found
    """
    for m in re.finditer(r"\${(\w+)}", input_str):
        var = m.group(1)
        try:
            input_str = re.sub(r'\${' + var + '}', config(var), input_str)
        except UndefinedValueError:
            raise ValueError(f"Invalid env var {m.group(1)} in {input_str}")
        
    return input_str
